SystemOrganization addCategory: #'LRSTM-Core-Compiler'!
SystemOrganization addCategory: #'LRSTM-Core-Errors'!
SystemOrganization addCategory: #'LRSTM-Core-Model'!
SystemOrganization addCategory: #'LRSTM-Core-OmniCompatibility'!

Array variableSubclass: #OstArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-OmniCompatibility'!

!Process methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/30/2007 12:09'!
currentTransaction
	^ currentTransaction! !

!Process methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/30/2007 12:09'!
currentTransaction: aTransaction
	currentTransaction := aTransaction! !

!BlockContext methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 2/26/2008 18:16'!
atomic
	^ LRSTMTransaction new do: self! !

!BlockContext methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:13'!
atomicIfConflict: aBlock
	^ LRSTMTransaction new do: self ifConflict: aBlock! !

!BlockContext methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:14'!
atomicRetry
	^ LRSTMTransaction new retry: self! !

TestCase subclass: #OstTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-OmniCompatibility'!

!BlockClosure methodsFor: '*LRSTM-Core-Runtime' stamp: 'StefanMarr 11/28/2011 20:58'!
atomic
	^ LRSTMTransaction new do: self! !

!BlockClosure methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:14'!
atomicIfConflict: aBlock
	^ LRSTMTransaction new do: self ifConflict: aBlock! !

!BlockClosure methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/25/2007 09:15'!
atomicRetry
	^ LRSTMTransaction new retry: self! !

RBCompiler subclass: #LRSTMCompiler
	instanceVariableNames: 'doFullTransformation'
	classVariableNames: 'ErrorOnMissingAtomicMethod OldCompiler'
	poolDictionaries: ''
	category: 'LRSTM-Core-Compiler'!

!LRSTMCompiler commentStamp: 'StefanMarr 11/28/2011 22:47' prior: 0!
This is just a helper class holding the compilation related methods similar to the original implementation of ACCompiler!

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 12/3/2011 23:13'!
atomicFail: aCompiledMethod
	"Failed to compile aCompiledMethod. Show error in transcript and install the origianl method."
	<atomicDoNotTransform>
	
	Transcript
		show: 'ERROR: Failed to compile for atomic execution: '; show: aCompiledMethod methodClass; 
		show: '>>#'; show: aCompiledMethod selector; cr.
	^ aCompiledMethod! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 12/4/2011 12:24'!
atomicMethod: aMessage missing: aReceiver
	"This method is called whenever a missing atomic method is called. Installs a whole hierarchy of atomic methods."
	<atomicDoNotTransform>
	^ self atomicMethod: aMessage missing: aReceiver in: aReceiver class ! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
atomicMethod: aMessage missing: aReceiver in: aClass
	"This method is called whenever a missing atomic method is called. Installs a whole hierarchy of atomic methods."
	<atomicDoNotTransform>
	
	| selector addedAtomicMethod curClass |
	selector := aMessage selector asNormalSelector.
	
	ErrorOnMissingAtomicMethod ifTrue: [
		^ (MessageNotUnderstood new
				message: aMessage;
				receiver: self;
				signal)].
	
	ProtoObject withAllSubclassesDo: [ :class |
		(class methodDictionary includesKey: selector) ifTrue: [
			self installAtomicVersionOf: (class methodDictionary at: selector).] ].
! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/5/2011 16:32'!
disableCompiler
	RBCompilerDispatcher current: OldCompiler.! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
disableErrorOnMissingAtomicMethod
	ErrorOnMissingAtomicMethod := false.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/5/2011 16:31'!
enableCompiler
	OldCompiler := RBCompilerDispatcher current.
	RBCompilerDispatcher current: self.! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
errorOnMissingAtomicMethod
	ErrorOnMissingAtomicMethod := true.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 22:05'!
evaluate: textOrString notifying: aController logged: logFlag 
	"This is used for DoIts, the do not seem to be supported currently by OPAL. Let's use a plain compiler instead."
	Transcript cr; show: 'Did not use the LRSTM compiler to compile the following code:'.
	Transcript cr; show: '---'.
	Transcript cr; show: textOrString contents.
	Transcript cr; show: '---'.
	
	^ Compiler evaluate: textOrString notifying: aController logged: logFlag.! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 9/12/2012 13:54'!
initialize
	Transcript cr; show: '===INITIALIZE LRSTM [START]===='.
	
	ErrorOnMissingAtomicMethod := false.
	
	(Process hasInstVarNamed: 'currentTransaction') ifFalse: [
		Process addInstVarNamed: 'currentTransaction'.
	].
	
	self recompileClassesWithLRSTMCompiler.
	
	SystemChangeNotifier uniqueInstance
		notify: self ofSystemChangesOfItem: #method change: #Added using: #onMethodAdded:;
		notify: self ofSystemChangesOfItem: #method change: #Modified using: #onMethodAdded:;
		notify: self ofSystemChangesOfItem: #method change: #Removed using: #onMethodRemoved:.
	
	Transcript cr; show: '===INITIALIZE LRSTM [COMPLETED]===='.
! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 12/3/2011 23:14'!
installAtomicVersionOf: aCompiledMethod
	"Answer the atomic counterpart of aCompiledMethod.
	Methods with the pragma #atomicDoNotTransform are considered to be already atomic.
	Methods with the pragma #atomic: will use the annotation code for the atomic counterpart.
	#atomicInstead: will use the annotiation code for atomic transformation."
	<atomicDoNotTransform>

	| atomicCM atomicName |
	"Is already atomic do nothing"
	aCompiledMethod selector isAtomicSelector ifTrue: [
		^ aCompiledMethod ].

	atomicName := aCompiledMethod selector asAtomicSelector.
	
	aCompiledMethod pragmas do: [ :each |
		each keyword = #atomicDoNotTransform
			ifTrue: [ atomicCM := aCompiledMethod.
						aCompiledMethod methodClass methodDict at: atomicName put: atomicCM.
						^ atomicCM ].
		each keyword = #atomic:
			ifTrue: [
				atomicCM := self installAtomicVersionOf: 
								(aCompiledMethod methodClass lookupSelector: each arguments first).
				aCompiledMethod methodClass methodDict at: atomicName put: atomicCM.
				^ atomicCM ].
	].
	
	
	[ | newMethod |
		newMethod := LRSTMCompiler recompileAtomically: aCompiledMethod.
		aCompiledMethod methodClass methodDict at: atomicName put: newMethod.
	  ^ newMethod] value.

		"ifError: [ :err :rcvr |
						Transcript log: 'FAILED to compile ', atomicName, ' error: ' , err asString, ' rcvr: ', rcvr asString. 
						self haltOnce.
						self atomicFail: aCompiledMethod ]."
	^ nil! !

!LRSTMCompiler class methodsFor: 'events' stamp: 'StefanMarr 12/3/2011 23:14'!
onMethodAdded: anObject
	<atomicDoNotTransform>
	
	"Ignore the already atomic methods"
	anObject itemSelector isAtomicSelector ifTrue: [
		^self].
	
	"Also blacklist a few things"
	((anObject itemClass = Debugger)
	or: [anObject itemClass = Decompiler])
		ifTrue: [^ self].

	"For the rest we do the atomic stuff"	
	self installAtomicVersionOf: (anObject itemClass methodDict
			at: anObject itemSelector).! !

!LRSTMCompiler class methodsFor: 'events' stamp: 'StefanMarr 12/3/2011 23:14'!
onMethodRemoved: anObject
	<atomicDoNotTransform>
	
	anObject itemClass methodDict 
		removeKey: anObject itemSelector 
			asAtomicSelector
		ifAbsent: nil! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 23:14'!
recompile: aCompiledMethod
	"Recompile the given method normally"
	<atomicDoNotTransform>
	
	| compiler |
	compiler := LRSTMCompiler new.
	^ compiler recompile: aCompiledMethod.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 23:14'!
recompileAtomically: aCompiledMethod
	"Recompile the given method normally"
	<atomicDoNotTransform>
	
	| compiler |
	compiler := LRSTMCompiler new.
	^ compiler recompileAtomically: aCompiledMethod.! !

!LRSTMCompiler class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/30/2011 20:40'!
recompileClassesWithLRSTMCompiler
	<atomicDoNotTransform>
	
	self enableCompiler.

	(PackageInfo named: 'LRSTM') classes
		do: [:c | c compileAll]
		displayingProgress: 'Recompiling with LRSTM support enabled'.! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/3/2011 23:16'!
unload
	Smalltalk allClassesAndTraits 
		do: [ :ea | ea flushAtomic. ea class flushAtomic ]
		displayingProgress: 'Flushing Atomic Methods'.
	Process removeInstVarNamed: 'currentTransaction'! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/2/2011 11:42'!
unregisterChangeListener

	SystemChangeNotifier uniqueInstance
		noMoreNotificationsFor: self.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 18:02'!
initialize
	super initialize.
	doFullTransformation := false.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:44'!
parse: methodBoolean
	| methodNode transformer |
	methodNode := super parse: methodBoolean.
	
	methodBoolean ifTrue: [
		transformer := LRSTMTransformer new.
		transformer fullTransformation: doFullTransformation.
		transformer applySTMTransformations: methodNode inScopeOf: class.
	].
	^ methodNode! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 18:06'!
recompile: aCompiledMethod
	"Recompile the given method normally, i.e., with out full STM transformation"

	^ self recompile: aCompiledMethod atomically: false.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:05'!
recompile: aCompiledMethod atomically: aBool
	"Recompile the given method with the STM enabled"

	| methodSource newSelector  newMethod |
	
	doFullTransformation := aBool.
	
	methodSource := aCompiledMethod getSource.
	newSelector := aCompiledMethod selector asAtomicSelector.
		
	newMethod := self
		compile: methodSource in: aCompiledMethod methodClass notifying: self
		ifFail: [^ self problem: 'syntax error' sel: newSelector in: aCompiledMethod methodClass].

	^ newMethod generate.! !

!LRSTMCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 18:06'!
recompileAtomically: aCompiledMethod
	"Recompile the given method with the STM enabled"

	^ self recompile: aCompiledMethod atomically: true.! !

RBSemanticAnnotator subclass: #LRSTMTransformer
	instanceVariableNames: 'doFullTransformation doLocalTransformation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Compiler'!

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:27'!
acceptAssignmentNode: anAssignmentNode
	super acceptAssignmentNode: anAssignmentNode.
	
	self doTransformation ifFalse: [^self].
	
	(anAssignmentNode variable isInstanceBinding or:
	[anAssignmentNode variable isLiteralBinding "or: REMOTE NOT YET SUPPORTED
	[anAssignmentNode variable isRemote]"]) ifTrue: [
		| replacement |
		replacement := self createAtomicAssignment: anAssignmentNode.
		anAssignmentNode parent replaceNode: anAssignmentNode withNode: replacement.
	]! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 18:50'!
acceptMessageNode: aMessageNode
	| doLocalTransformationOld |
	doLocalTransformationOld := doLocalTransformation.
	
	doLocalTransformation := doLocalTransformation or: 
								 [aMessageNode selector = #atomic or:
								 [aMessageNode selector = #atomicIfConflict: or:
								 [aMessageNode selector = #atomicRetry           ]]].

	(self doTransformation and: [aMessageNode isInlined not])
		ifTrue:  [self  transformMessageNode: aMessageNode.]
		ifFalse: [super acceptMessageNode:    aMessageNode.].
	
	doLocalTransformation := doLocalTransformationOld.! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:55'!
acceptMethodNode: aMethodNode
	super acceptMethodNode: aMethodNode.
	doFullTransformation ifTrue: [
		aMethodNode selector: (aMethodNode selector asAtomicSelector)
	].! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:31'!
acceptVariableNode: aVariableNode
	| replacement |
	self doTransformation ifFalse: [^self].
	super acceptVariableNode: aVariableNode.
	
	"Not every variable node we are going to encounter is a read.
	 Make sure it is not the variable node of a parent assignment node.
	 Furthermore, we do not need to take care of temp assignments.
	 Only globals, instances, and perhaps remote onces of a closure."
	(aVariableNode isRead
	 and: [aVariableNode isLiteralBinding   or: 
		   [aVariableNode isInstanceBinding "or: REMOTES NOT SUPPORTED at the moment
			[aVariableNode isRemote]"]])
		ifFalse: [^self].
	
	"Now we are sure that we have to transform a variable read".
	replacement := self createAtomicReadNode: aVariableNode.
	aVariableNode parent replaceNode: aVariableNode withNode: replacement. 
! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:10'!
applySTMTransformations: methodNode inScopeOf: aClass
	"We have two cases:
	  Either it is a normal method, and we do only the atomic blocks,
	  or it is a fully atomic method and we transform everything.
	
		doFullTransformation indicates that we do not only local adaptations.
	"
	| classScope |
	(self pragmasForbidTransformation: methodNode)
		ifTrue: [^self].
	
	"methodNode acceptVisitor: self."
	classScope :=  (RBVariableScope 
		owner: (RBLiteralScope 
			owner: RBRootScope new 
			class: aClass)
		class: aClass).
	
	self start: methodNode scope: classScope.! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:47'!
createAtomicAssignment: anAssignmentNode
	| var |
	var := anAssignmentNode variable.
	var isInstanceBinding ifTrue: [
		| selfRef instIdx sendArgs |
		selfRef := RBVariableNode named: 'self'.
		selfRef variableBinding: (var variableBinding scope lookup: 'self').
		
		instIdx := RBLiteralValueNode value: (var variableBinding index).
		
		sendArgs := OrderedCollection new: 2.
		sendArgs add: instIdx.
		sendArgs add: anAssignmentNode value.
		^ RBMessageNode receiver: selfRef selector: #atomicInstVarAt:put: arguments: sendArgs.
	]
	ifFalse: [
		| globalAssoc sendArgs |
		var isLiteralBinding ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		
		globalAssoc := RBLiteralValueNode value: var variableBinding binding.
		
		sendArgs := OrderedCollection new: 1.
		sendArgs add: anAssignmentNode value.
		^ RBMessageNode receiver: globalAssoc selector: #atomicValue: arguments: sendArgs.
	].! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 15:50'!
createAtomicReadNode: aVariableNode
	| selfRef instIdx sendArgs |
	
	aVariableNode isInstanceBinding ifTrue: [
		selfRef := RBVariableNode named: 'self'.
		selfRef variableBinding: (aVariableNode variableBinding scope lookup: 'self').
		
		instIdx := RBLiteralValueNode value: (aVariableNode variableBinding index).
		sendArgs:= OrderedCollection new: 1.
		sendArgs add: instIdx.
		^ RBMessageNode receiver: selfRef selector: #atomicInstVarAt: arguments: sendArgs.	
	]
	ifFalse: [
		| globalAssoc |
		aVariableNode isLiteralBinding ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		
		globalAssoc := RBLiteralValueNode value: aVariableNode variableBinding binding.
		^ RBMessageNode receiver: globalAssoc selector: #atomicValue.
	].
	! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:08'!
doTransformation
	^ doFullTransformation or: [doLocalTransformation]! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:56'!
fullTransformation: aBool
	doFullTransformation := aBool! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/4/2011 14:52'!
initialize
	super initialize.
	doFullTransformation  := false.
	doLocalTransformation := false.
! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 23:50'!
pragmasForbidTransformation: methodNode
	methodNode pragmas do: [:each |
		(each selector = #atomicDoNotTransform or: [each selector = #atomic:])
			ifTrue: [^true].
	].
	^false.! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:07'!
transformMessageNode: aMessageNode

	| selector |
	selector := aMessageNode selector.
	(selector = #atomic or: [aMessageNode isInlined]) ifFalse: [
		aMessageNode selector: selector asAtomicSelector.
	].
	
	super acceptMessageNode: aMessageNode.! !

!ContextPart methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/31/2007 15:49'!
workingCopy
	^ self! !

Error subclass: #LRSTMConflict
	instanceVariableNames: 'transaction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Errors'!

!LRSTMConflict methodsFor: 'accessing' stamp: 'lr 5/25/2007 09:04'!
transaction
	^ transaction! !

!LRSTMConflict methodsFor: 'accessing' stamp: 'lr 5/25/2007 09:04'!
transaction: aTransaction
	transaction := aTransaction! !

Error subclass: #LRSTMInvalidTransaction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Errors'!

Object subclass: #LRSTMChange
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Model'!

!LRSTMChange methodsFor: 'actions' stamp: 'lr 6/8/2007 16:59'!
apply
	self subclassResponsibility! !

!LRSTMChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasChanged
	^ true! !

!LRSTMChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasConflict
	^ false! !

LRSTMChange subclass: #LRSTMCustomChange
	instanceVariableNames: 'applyBlock conflictTestBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Model'!

!LRSTMCustomChange class methodsFor: 'instance-creation' stamp: 'StefanMarr 12/4/2011 20:08'!
onApply: anApplyBlock
	<atomicDoNotTransform>
	^ self onApply: anApplyBlock onConflictTest: [ false ]! !

!LRSTMCustomChange class methodsFor: 'instance-creation' stamp: 'StefanMarr 12/4/2011 20:09'!
onApply: anApplyBlock onConflictTest: aConflictTestBlock
	<atomicDoNotTransform>
	^ self basicNew initializeApply: anApplyBlock conflictTest: aConflictTestBlock! !

!LRSTMCustomChange methodsFor: 'actions' stamp: 'lr 2/26/2008 16:26'!
apply
	applyBlock value! !

!LRSTMCustomChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:28'!
hasConflict
	^ conflictTestBlock value! !

!LRSTMCustomChange methodsFor: 'initialization' stamp: 'lr 6/8/2007 17:04'!
initializeApply: anApplyBlock conflictTest: aConflictTestBlock	
	applyBlock := anApplyBlock.
	conflictTestBlock := aConflictTestBlock! !

LRSTMChange subclass: #LRSTMObjectChange
	instanceVariableNames: 'original previous working'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Model'!

!LRSTMObjectChange class methodsFor: 'instance-creation' stamp: 'lr 6/8/2007 16:59'!
on: anObject
	^ self basicNew initializeOn: anObject! !

!LRSTMObjectChange methodsFor: 'actions' stamp: 'lr 6/8/2007 16:59'!
apply
	original restoreSnapshot: working! !

!LRSTMObjectChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasChanged
	^ (working isIdenticalToSnapshot: previous) not! !

!LRSTMObjectChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasConflict
	^ (original isIdenticalToSnapshot: previous) not! !

!LRSTMObjectChange methodsFor: 'initialization' stamp: 'lr 6/8/2007 16:59'!
initializeOn: anObject
	original := anObject.
	working := original snapshotCopy.
	previous := working snapshotCopy! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
original
	^ original! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
previous
	^ previous! !

!LRSTMObjectChange methodsFor: 'printing' stamp: 'lr 6/8/2007 17:01'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' ['; print: self original; nextPut: $]! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
working
	^ working! !

Object subclass: #LRSTMTransaction
	instanceVariableNames: 'parent context changes objectChanges'
	classVariableNames: ''
	poolDictionaries: 'ContextPart'
	category: 'LRSTM-Core-Model'!

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:45'!
abort
	"Abort a transaction."
	
	<atomicDoNotTransform>
	parent := context := changes := objectChanges := nil! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 18:32'!
abort: aValue
	"Abort a transaction."
	
	<atomicDoNotTransform>
	Processor activeProcess currentTransaction: parent.
	thisContext swapSender: context.
	self abort.
	^ aValue! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 17:07'!
addChange: aChange
	<atomicDoNotTransform>
	^ changes add: aChange! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:42'!
begin
	"Start a transaction. If this is a nested transaction copy the parent changes into our own context."
	
	<atomicDoNotTransform>
	changes := OrderedCollection new.
	objectChanges := IdentityDictionary new.
	self isNested ifTrue: [
		parent objectChanges
			keysAndValuesDo: [ :key :value | objectChanges at: key put: value copy ] ]! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 17:07'!
changeFor: anObject
	<atomicDoNotTransform>
	^ objectChanges at: anObject ifAbsentPut: [ self addChange: (LRSTMObjectChange on: anObject) ]! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 17:07'!
changes
	"Answer an ordered collection of unique changes."

	<atomicDoNotTransform>
	^ changes! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 16:59'!
checkpoint
	"Checkpoint a transaction."

	<atomicDoNotTransform>
	self commit; begin! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 6/7/2007 16:47'!
commit
	"Commit a transaction."

	<atomicDoNotTransform>
	self commitIfConflict: [ self signalConflict ]! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:46'!
commitIfConflict: aBlock
	"Commit a transaction atomically."
	
	<atomicDoNotTransform>
	[	changes do: [ :each |
			each hasConflict
				ifTrue: [ ^ aBlock value ] ].
		self isNested 
			ifTrue: [
				objectChanges keysAndValuesDo: [ :key :value |
					(parent objectChanges includesKey: key)
						ifFalse: [ parent addChange: value ].
					parent objectChanges at: key put: value ] ]
			ifFalse: [ 	
				changes do: [ :each |
					each hasChanged
						ifTrue: [ each apply ] ] ] ]
		valueUnpreemptively! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 16:59'!
do: aBlock
	"Start a transaction with aBlock."

	<atomicDoNotTransform>		
	^ self do: aBlock ifConflict: [ self signalConflict ]! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 2/26/2008 18:38'!
do: aBlock ifConflict: aConflictBlock
	"Evaluate aBLock within a new transaction, unless we are already in an existing transactional context. Evaluate aConflictBlock if the transaction conflicts with concurrent edits."

	<atomicDoNotTransform>
	| result |
	parent := Processor activeProcess currentTransaction.
	Processor activeProcess currentTransaction: self.
	context := thisContext sender.
	self begin.
	[ result := aBlock 
		ensure: [ Processor activeProcess currentTransaction: parent ] ]
		ifCurtailed: [ self abort ].
	self commitIfConflict: aConflictBlock.
	^ result! !

!LRSTMTransaction methodsFor: 'testing' stamp: 'lr 2/26/2008 18:42'!
isNested
	<atomicDoNotTransform>
	^ parent notNil! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 18:44'!
objectChanges
	<atomicDoNotTransform>
	^ objectChanges! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'lr 2/26/2008 18:31'!
parent
	<atomicDoNotTransform>
	^ parent! !

!LRSTMTransaction methodsFor: 'public' stamp: 'lr 6/7/2007 12:10'!
retry: aBlock
	"Unconditionally retry to evaluate aBlock until there are no conflict."

	<atomicDoNotTransform>
	^ self do: aBlock ifConflict: [ self retry: aBlock ]! !

!LRSTMTransaction methodsFor: 'private' stamp: 'lr 2/26/2008 18:09'!
signalConflict
	<atomicDoNotTransform>
	^ LRSTMConflict new transaction: self; signal! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 6/7/2007 15:05'!
isIdenticalToSnapshot: anObject
	1 to: self class instSize do: [ :index | 
		(self instVarAt: index) == (anObject instVarAt: index)
			ifFalse: [ ^ false ] ].
	self class isVariable ifTrue: [
		1 to: self size do: [ :index | 
			(self at: index) == (anObject at: index)
				ifFalse: [ ^ false ] ] ].
	^ true! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 6/7/2007 13:47'!
restoreSnapshot: anObject
	<primitive: 168>
	self primitiveFailed! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 6/7/2007 13:48'!
snapshotCopy
	<primitive: 148>
	self primitiveFailed! !

!Object methodsFor: '*LRSTM-Core-Runtime' stamp: 'lr 5/30/2007 12:11'!
workingCopy
	"Answer a working copy to be used within the context of the current transaction."
	
	| transaction |
	transaction := Processor activeProcess currentTransaction
		ifNil: [ self error: 'No active transaction' ].
	^ (transaction changeFor: self) working! !

Object subclass: #OstObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-OmniCompatibility'!

!OstObject class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/15/2012 22:35'!
omnifyClass: cls! !
LRSTMCompiler initialize!
